<?php

namespace Drupal\latvia_auth_saml_example;

use Drupal\Component\Utility\Crypt;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Site\Settings;
use Drupal\latvia_auth_saml\AuthenticationService;
use Drupal\latvia_auth_saml\UserData;
use Drupal\latvia_auth_saml_example\Exception\LoginException;
use Drupal\user\Entity\User;
use Psr\Log\LoggerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Custom login authentication service implementation.
 */
class ExampleLoginAuthenticationService extends AuthenticationService {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The example authentication service constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger instance.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, LoggerInterface $logger) {
    $this->entityTypeManager = $entity_type_manager;
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   */
  public function processLogin(array $attributes, $return_to = NULL) {
    $user_data = UserData::create($attributes);
    $national_id = $user_data->getFormattedNationalIdNumber();

    try {
      // Get the hashed account name.
      $hashed_account_name = $this->getHashedAccountName($national_id);

      // Load or create the account based on the national ID.
      if (!($account = $this->load($hashed_account_name))) {
        $account = User::create([
          'name' => $hashed_account_name,
        ]);
        $account->save();
      }

      // Log in the user.
      user_login_finalize($account);

      $this->logger->notice('User %name has successfully logged in via Latvija.lv.', ['%name' => $account->getAccountName()]);
    }
    catch (\Exception $e) {
      watchdog_exception('latvia_auth_saml_example', $e);

      throw new LoginException('The user could not be logged in.');
    }

    return new RedirectResponse($return_to);
  }

  /**
   * Returns the hashed account name.
   *
   * @param string $national_id
   *   The national ID.
   *
   * @return string
   *   The hashed account name.
   */
  protected function getHashedAccountName($national_id) {
    $key = Settings::get('latvia_auth_saml.encryption.key');

    return Crypt::hmacBase64($national_id, $key);
  }

  /**
   * Load a Drupal user based on an external identifier.
   *
   * @param string $username
   *   The account username.
   *
   * @return \Drupal\user\UserInterface|null
   *   The loaded Drupal user.
   */
  protected function load($username) {
    $user_storage = $this->entityTypeManager->getStorage('user');
    $users = $user_storage->loadByProperties(['name' => $username]);

    if ($users && count($users) == 1) {
      return reset($users);
    }

    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function processLogout() {
    user_logout();
  }

}
