<?php

namespace Drupal\latvia_auth_saml_example;

use Drupal\Core\Http\RequestStack;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\latvia_auth_saml\AuthenticationService as OriginalAuthenticationService;
use Drupal\latvia_auth_saml\Controller\AuthController;
use Drupal\latvia_auth_saml\Exception\UserDataParsingException;
use Drupal\latvia_auth_saml\FilterInterface;
use Drupal\latvia_auth_saml\UserData;
use Psr\Log\LoggerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Custom authentication service implementation.
 */
class ExampleSessionAuthenticationService extends OriginalAuthenticationService {

  use StringTranslationTrait;

  /**
   * Defines the guest data variable name in session storage.
   */
  const SESSION_USER_DATA = 'user_data';

  /**
   * The encryption filter service instance.
   *
   * @var \Drupal\latvia_auth_saml\FilterInterface
   */
  protected $encryptionFilter;

  /**
   * Current request.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $request;

  /**
   * The logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * Authentication handler constructor.
   *
   * @param \Drupal\latvia_auth_saml\FilterInterface $encryption_filter
   *   The encryption filter service instance.
   * @param \Drupal\Core\Http\RequestStack $request_stack
   *   Request stack.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger instance.
   */
  public function __construct(FilterInterface $encryption_filter, RequestStack $request_stack, LoggerInterface $logger) {
    $this->encryptionFilter = $encryption_filter;
    $this->request = $request_stack->getCurrentRequest();
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   */
  public function processLogin(array $attributes, $return_to = NULL) {
    if (!empty($attributes)) {
      $session = $this->request->getSession();
      $session->migrate();

      try {
        // Encode user data.
        $user_data_encoded = UserData::toJson($attributes, $this->encryptionFilter);

        // Store user data in the session.
        $session->set(self::SESSION_USER_DATA, $user_data_encoded);

        // Log an event.
        $user_data = UserData::create($attributes);
        $this->logger->notice('User with given name %name has successfully logged in via Latvija.lv.', ['%name' => $user_data->getGivenName()]);

        // Get return to URL.
        $return_to = AuthController::validateReturnUrl($return_to);

        return new RedirectResponse($return_to);
      }
      catch (\Exception $e) {
        watchdog_exception('latvia_auth_saml_example', $e);

        throw new UserDataParsingException($this->t('User data could not be stored.'));
      }
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function processLogout() {
    $session = $this->request->getSession();

    // Invalidate the session.
    $session->remove(self::SESSION_USER_DATA);
  }

}
