<?php

namespace Drupal\latvia_auth_saml\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Render\RenderContext;
use Drupal\Core\Routing\TrustedRedirectResponse;
use Drupal\Core\Site\Settings;
use Drupal\Core\Url;
use Drupal\latvia_auth_saml\AuthenticationServiceInterface;
use Drupal\latvia_auth_saml\DevelopmentLogger;
use Drupal\latvia_auth_saml\Exception\PassThroughExceptionMessageInterface;
use Drupal\latvia_auth_saml\Exception\SamlResponseMissingException;
use Drupal\latvia_auth_saml\Exception\SamlResponseNotValidException;
use Drupal\latvia_auth_saml\SamlSettings;
use OneLogin\Saml2\Auth;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Session\SessionInterface;

/**
 * The authentication controller.
 */
class AuthController extends ControllerBase {

  /**
   * Saml settings service instance.
   *
   * @var \Drupal\latvia_auth_saml\SamlSettings
   */
  protected $samlSettings;

  /**
   * The authentication service instance.
   *
   * @var \Drupal\latvia_auth_saml\AuthenticationServiceInterface
   */
  protected $authenticationService;

  /**
   * The session instance.
   *
   * @var \Symfony\Component\HttpFoundation\Session\SessionInterface
   */
  protected $session;

  /**
   * The logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * Default redirect route name.
   *
   * @var string
   */
  protected $defaultRedirectRouteName = '<front>';

  /**
   * Authentication controller constructor.
   *
   * @param \Drupal\latvia_auth_saml\SamlSettings $saml_settings
   *   The Saml settings service.
   * @param \Drupal\latvia_auth_saml\AuthenticationServiceInterface $authentication_service
   *   The authentication service.
   * @param \Symfony\Component\HttpFoundation\Session\SessionInterface $session
   *   The session instance.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger instance.
   */
  public function __construct(SamlSettings $saml_settings, AuthenticationServiceInterface $authentication_service, SessionInterface $session, LoggerInterface $logger) {
    $this->samlSettings = $saml_settings;
    $this->authenticationService = $authentication_service;
    $this->session = $session;
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $authentication_service = Settings::get(
      // Authentication service name defined in settings.php.
      'latvia_auth_saml.authentication_service_name',
      // The default authentication service.
      'latvia_auth_saml.authentication_service'
    );

    return new static(
      $container->get('latvia_auth_saml.saml_settings'),
      $container->get($authentication_service),
      $container->get('session'),
      $container->get('logger.channel.latvia_auth_saml')
    );
  }

  /**
   * Initiates a single sign-on request.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   Request instance.
   *
   * @return \Symfony\Component\HttpFoundation\Response|string
   *   Response object.
   */
  public function singleSignOn(Request $request) {
    try {
      $render_context = new RenderContext();

      // Get response URL in render context. Without the context, Drupal would
      // throw an error about leaked cache metadata.
      $response_url = \Drupal::service('renderer')->executeInRenderContext($render_context, function () use ($request) {
        // Log the event.
        $this->logger->info(t('Single sign-on request is initiated.'));

        $saml_settings = $this->samlSettings->get();
        $auth = new Auth($saml_settings);

        // Get query params.
        $params = $request->query->all();
        // Validate return-to URL.
        $return_to = static::validateReturnUrl($params['returnTo'] ?? NULL);

        // Prepare query parameters.
        $query_params = [
          'culture' => $this->languageManager()->getCurrentLanguage()->getId(),
        ];

        // Get request URL.
        $request_url = $auth->login($return_to ?: NULL, $query_params, FALSE, FALSE, TRUE);

        // Set request ID in the session.
        $this->session->set(AuthenticationServiceInterface::SESSION_AUTH_REQUEST_ID, $auth->getLastRequestID());

        return $request_url;
      });

      // Prepare response.
      $response = new TrustedRedirectResponse($response_url);
      $response->addCacheableDependency($render_context);

      return $response;
    }
    catch (\Exception $e) {
      // Prepare error message.
      $t_message = $e instanceof PassThroughExceptionMessageInterface ? $e->getMessage() : t('Authentication failed. Please contact site administrator.');
      $this->messenger()->addError($t_message);
      // Log the error.
      DevelopmentLogger::logException($e);
    }

    return $this->redirect($this->defaultRedirectRouteName);
  }

  /**
   * Processes Saml assertion response.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   Current request.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Response object.
   */
  public function assertionConsumerService(Request $request) {
    // Proceed with sign-on assertion in case of POST request.
    if ($request->getMethod() == 'POST') {
      try {
        if (isset($_POST['SAMLResponse'])) {
          // Log the event.
          $this->logger->info(t('Assertion response processing is initiated.'));

          $saml_settings = $this->samlSettings->get();
          $auth = new Auth($saml_settings);

          // Get request ID from the session.
          $request_id = $this->session->get('authn_request_id');

          // Process response.
          $auth->processResponse($request_id);

          // In case of errors, thrown a debuggable exception.
          if ($auth->getErrors()) {
            throw new SamlResponseNotValidException([
              $auth->getLastErrorReason(),
            ]);
          }

          // Unset request ID from the session.
          $this->session->remove(AuthenticationServiceInterface::SESSION_AUTH_REQUEST_ID);

          // Get attributes.
          $attributes = $auth->getAttributes();

          // Log the event.
          $this->logger->info(t('Assertion response processing is completed.'));

          // Authenticate the user.
          if ($redirect_response = $this->authenticationService->processLogin($attributes, $_POST['RelayState'] ?? NULL)) {
            return $redirect_response;
          }
        }
        else {
          throw new SamlResponseMissingException();
        }
      }
      catch (\Exception $e) {
        // Prepare error message.
        $t_message = $e instanceof PassThroughExceptionMessageInterface ? $e->getMessage() : t('Authentication failed. Please contact site administrator.');
        $this->messenger()->addError($t_message);
        // Log the error.
        DevelopmentLogger::logException($e);
      }
    }
    // Pass GET requests to single logout service.
    elseif ($request->getMethod() == 'GET') {
      return $this->singleLogoutService($request);
    }

    return $this->redirect($this->defaultRedirectRouteName);
  }

  /**
   * Initiates a single sign-out request.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Response object.
   */
  public function singleLogout() {
    try {
      $render_context = new RenderContext();

      // Get response URL in render context. Without the context, Drupal would
      // throw an error about leaked cache metadata.
      $response_url = \Drupal::service('renderer')->executeInRenderContext($render_context, function () {
        // Log the event.
        $this->logger->info(t('Single log-out SP-initiated request is initiated.'));

        $saml_settings = $this->samlSettings->get();
        $auth = new Auth($saml_settings);

        // Prepare return-to URL.
        $return_to = Url::fromRoute($this->defaultRedirectRouteName)->setAbsolute()->toString();

        // Prepare query parameters.
        $query_params = [
          'culture' => $this->languageManager()->getCurrentLanguage()->getId(),
        ];

        // Get request URL.
        $request_url = $auth->logout($return_to, $query_params, NULL, NULL, TRUE);

        // Set request ID in the session.
        $this->session->set(AuthenticationServiceInterface::SESSION_LOGOUT_REQUEST_ID, $auth->getLastRequestID());

        // Log the event.
        $this->logger->info(t('Single log-out SP-initiated request is completed.'));

        return $request_url;
      });

      // Prepare response.
      $response = new TrustedRedirectResponse($response_url);
      $response->addCacheableDependency($render_context);

      return $response;
    }
    catch (\Exception $e) {
      $t_message = $e instanceof PassThroughExceptionMessageInterface ? $e->getMessage() : t('Log-out failed. Please contact site administrator.');
      $this->messenger()->addError($t_message);
      // Log the error.
      DevelopmentLogger::logException($e);

      return $this->redirect($this->defaultRedirectRouteName);
    }
  }

  /**
   * Processes Saml single sign-out response.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Response object.
   */
  public function singleLogoutService(Request $request) {
    try {
      $render_context = new RenderContext();
      $action_context = NULL;

      // Determine logout action context.
      if (isset($_GET['SAMLRequest'])) {
        $action_context = 'request';
      }
      elseif (isset($_GET['SAMLResponse'])) {
        $action_context = 'response';
      }

      $response_url = \Drupal::service('renderer')->executeInRenderContext($render_context, function () use ($action_context) {
        switch ($action_context) {
          case 'request':
            $this->logger->info(t('Single log-out IdP-initiated request is initiated.'));
            break;

          case 'response':
            $this->logger->info(t('Single log-out IdP-initiated response is initiated.'));
            break;

        }

        $saml_settings = $this->samlSettings->get();
        $auth = new Auth($saml_settings);

        // Get request ID from the session.
        $request_id = $this->session->get(AuthenticationServiceInterface::SESSION_LOGOUT_REQUEST_ID);

        // Process log-out request.
        $callback = [$this->authenticationService, 'processLogout'];
        $response_url = $auth->processSLO(FALSE, $request_id, FALSE, $callback, TRUE);

        if ($auth->getErrors()) {
          throw new SamlResponseNotValidException([
            $auth->getLastErrorReason(),
          ]);
        }

        // Unset request ID from the session.
        $this->session->remove(AuthenticationServiceInterface::SESSION_LOGOUT_REQUEST_ID);

        switch ($action_context) {
          case 'request':
            $this->logger->info(t('Single log-out IdP-initiated request is completed.'));
            break;

          case 'response':
            $this->logger->info(t('Single log-out IdP-initiated response is completed.'));
            break;

        }

        return $response_url;
      });

      // Prepare response.
      if ($response_url) {
        $response = new TrustedRedirectResponse($response_url);
        $response->addCacheableDependency($render_context);

        return $response;
      }
    }
    catch (\Exception $e) {
      $t_message = $e instanceof PassThroughExceptionMessageInterface ? $e->getMessage() : t('Log-out failed. Please contact site administrator.');
      $this->messenger()->addError($t_message);
      // Log the error.
      DevelopmentLogger::logException($e);
    }

    return $this->redirect($this->defaultRedirectRouteName);
  }

  /**
   * Validates return URL.
   *
   * If the return URL is not valid or does not match the host, the URL to the
   * frontpage is returned.
   *
   * @param string $url
   *   Return-to URL.
   *
   * @return string|false
   *   Return URL.
   */
  public static function validateReturnUrl($url) {
    if ($url) {
      $host = \Drupal::request()->getHost();

      if ($parsed_url = parse_url($url)) {
        $parsed_host = $parsed_url['host'] ?? NULL;

        if ($host == $parsed_host) {
          return $url;
        }
      }
    }

    return Url::fromRoute('<front>')->setAbsolute()->toString();
  }

}
