<?php

namespace Drupal\latvia_auth_saml;

use Drupal\latvia_auth_saml\Exception\UserDataParsingException;

/**
 * User data wrapper.
 */
class UserData implements UserDataInterface {

  /**
   * The user data values.
   *
   * @var array
   *   The user data values.
   */
  protected $values = [];

  /**
   * Creates an instance of self from the given array.
   *
   * @param array $values
   *   The array with the user data.
   *
   * @return static
   *   Instance of self.
   */
  public static function create(array $values) {
    $result = new static();

    foreach ($values as $key => $value) {
      $result->set($key, $value);
    }

    return $result;
  }

  /**
   * Creates an instance of self from the given JSON string.
   *
   * @param string $json
   *   The JSON string with user data.
   * @param \Drupal\latvia_auth_saml\FilterInterface|null $filter
   *   The filter instance.
   *
   * @return static
   *   Instance of self.
   *
   * @throws \Drupal\latvia_auth_saml\Exception\UserDataParsingException
   *   An exception thrown if provided string is not a valid JSON string.
   * @throws \Exception
   *   An exception may be thrown by the filter instance.
   */
  public static function createFromJson($json, FilterInterface $filter = NULL) {
    $json = $filter ? $filter->unfilter($json) : $json;
    $values = json_decode($json);

    if (is_array($values) || is_object($values)) {
      $result = new static();

      foreach ($values as $key => $value) {
        $result->set($key, $value);
      }

      return $result;
    }

    throw new UserDataParsingException('The provided user data JSON string is not valid.');
  }

  /**
   * Sets the user data value.
   *
   * @param string $key
   *   Attribute name.
   * @param mixed $value
   *   Attribute value.
   */
  public function set($key, $value) {
    $this->values[$key] = serialize($value);
  }

  /**
   * Returns the requested value.
   *
   * @param string $key
   *   The attribute name.
   * @param mixed|null $default
   *   Default value.
   *
   * @return mixed
   *   Value of the attribute.
   */
  public function get($key, $default = NULL) {
    $values = @unserialize($this->values[$key]);

    return $values[0] ?? $default;
  }

  /**
   * {@inheritdoc}
   */
  public function getNationalIdNumber() {
    return $this->get('http://schemas.xmlsoap.org/ws/2005/05/identity/claims/privatepersonalidentifier');
  }

  /**
   * {@inheritdoc}
   */
  public function getFormattedNationalIdNumber() {
    $national_id = $this->getNationalIdNumber();
    // Offset of the first part of the ID number.
    $offset = 6;

    return sprintf('%s-%s', substr($national_id, 0, $offset), substr($national_id, $offset));
  }

  /**
   * {@inheritdoc}
   */
  public function getGivenName() {
    return mb_strtoupper($this->get('http://schemas.xmlsoap.org/ws/2005/05/identity/claims/givenname'));
  }

  /**
   * {@inheritdoc}
   */
  public function getSurname() {
    return mb_strtoupper($this->get('http://schemas.xmlsoap.org/ws/2005/05/identity/claims/surname'));
  }

  /**
   * Returns the user data values as JSON string (plain or encoded).
   *
   * @param array $values
   *   User data values.
   * @param \Drupal\latvia_auth_saml\FilterInterface|null $filter
   *   The encoding filter instance.
   *
   * @return string|false
   *   A JSON string (plain or encoded) or FALSE.
   *
   * @throws \Exception
   *   An exception may be thrown by the filter instance.
   */
  public static function toJson(array $values, FilterInterface $filter = NULL) {
    $json = json_encode($values);

    if ($json !== FALSE) {
      return $filter ? $filter->filter($json) : $json;
    }

    return FALSE;
  }

}
